(* file: svm.ml
 * author: Bob Muller
 * date: October 26, 2014, adapted from svm.py
 * revised: March 2015, added Mov, Mul and Div instructions.
 * revised: March 2016, removed Cpz; added 
 *                                   Li  (load immediate)
 *                                   Jsr (jump subroutine)
 *                                   R (return),
 *                      also added Zero, and Ra registers.
 *
 *
 * This program implements the Simple Virtual Machine (SVM). The
 * SVM has:
 *  - 16 instructions,
 *  -  8 registers, of these,
 *      - 4 are general purpose (R0 - R3)
 *      - 4 are special purpose:
 *       - pc   = program counter,
 *       - psw  = program status word,
 *       - Zero = contains 0
 *       - ra   = return address.
 *)

(* A type for the register file. The SVM has 4 general purpose registers.
*)
type registers = {r0:int; r1:int; r2:int; r3 : int};;

let string_of_regs {r0; r1; r2; r3} =
  let r0s = string_of_int r0 in
  let r1s = string_of_int r1 in
  let r2s = string_of_int r2 in
  let r3s = string_of_int r3
  in
  "{r0=" ^ r0s ^ "; r1=" ^ r1s ^ "; r2=" ^ r2s ^ "; r3=" ^ r3s ^ "}"

(* Names for registers.
*)
type reg = R0 | R1 | R2 | R3 | Zero 

let string_of_reg r = 
  match r with
  | R0 -> "R0" | R1 -> "R1" | R2 -> "R2" | R3 -> "R3" | Zero -> "Zero"

let registerGet register {r0; r1; r2; r3} =
  match register with
  | R0 -> r0 | R1 -> r1 | R2 -> r2 | R3 -> r3 | Zero -> 0

let registerPut value dest regs =
  match dest with
  | R0 -> { regs with r0=value}
  | R1 -> { regs with r1=value}
  | R2 -> { regs with r2=value}
  | R3 -> { regs with r3=value}
  | Zero -> failwith "registerPut: cannot write to the Zero register!"

(* These opnd types to satisfy OCaml's requirement for type declarations
 * for record types. 
 *)
type opnd0 = {rd:reg; rs:reg}
type opnd1 = {rd:reg; offset:int; rs:reg}
type opnd2 = {rs:reg; offset:int; rt:reg}
type opnd3 = {rd:reg; rs:reg; rt:reg}
type opnd4 = {rs:reg; rt:reg}
type opnd5 = {rd:reg; number:int}

type instruction = 
  | Lod of opnd1  (* {rd:reg; offset:int; rs:reg} *)
  | Li  of opnd5  (* {rd:reg; number:int} *)        
  | Sto of opnd2  (* {rs:reg; offset:int; rt:reg} *)
  | Mov of opnd0  (* {rd:reg; rs:reg} *)
  | Add of opnd3  (* {rd:reg; rs:reg; rt:reg} *)
  | Sub of opnd3  (* ditto *)
  | Mul of opnd3  (* ditto *)
  | Div of opnd3  (* ditto *)
  | Cmp of opnd4  (* {rs:reg; rt:reg } *)
  | Blt of int 
  | Beq of int
  | Bgt of int
  | Jmp of int
  | Jsr of int
  | R
  | Hlt

let string_of_instruction i =
  let si = string_of_int in
  let sr = string_of_reg in
  match i with
  | Lod {rd; offset; rs} ->
     "Lod\t" ^ (sr rd) ^ ", " ^ (si offset) ^ "(" ^ (sr rs) ^ ")"
  | Li  {rd; number} -> "Li\t" ^ (sr rd) ^ ", " ^ (si number)
  | Sto {rs; offset; rt} -> 
     "Sto\t" ^ (sr rs) ^ ", " ^ (si offset) ^ "(" ^ (sr rt) ^ ")"
  | Mov {rd; rs} -> "Mov\t" ^ (sr rd) ^ ", " ^ (sr rs)
  | Add {rd; rs; rt} -> "Add\t" ^ (sr rd) ^ ", " ^ (sr rs) ^ ", " ^ (sr rt)
  | Sub {rd; rs; rt} -> "Sub\t" ^ (sr rd) ^ ", " ^ (sr rs) ^ ", " ^ (sr rt)
  | Mul {rd; rs; rt} -> "Mul\t" ^ (sr rd) ^ ", " ^ (sr rs) ^ ", " ^ (sr rt)
  | Div {rd; rs; rt} -> "Div\t" ^ (sr rd) ^ ", " ^ (sr rs) ^ ", " ^ (sr rt)
  | Cmp {rs; rt} -> "Cmp\t" ^ (sr rs) ^ ", " ^ (sr rt)
  | Blt disp -> "Blt\t" ^ (si disp)
  | Beq disp -> "Beq\t" ^ (si disp)
  | Bgt disp -> "Bgt\t" ^ (si disp)
  | Jmp disp -> "Jmp\t" ^ (si disp)
  | Jsr disp -> "Jsr\t" ^ (si disp)
  | R -> "R"
  | Hlt -> "Hlt"
  
(* Object files have data segments and text segments. Representing 
 * segments as lists here. Pretty inefficient!
 *)
type datasegment = int list
type textsegment = instruction list

type image = {data:datasegment; text:textsegment}

let split segment n =
  let rec loop n xs ys =
    match (n, xs) with
    | (0, _) -> (List.rev ys, xs)
    | (n, z::zs) -> loop (n - 1) zs (z::ys)
    | _ -> failwith "cannot happen"
  in loop n segment [];;

let ramGet n segment = List.nth segment n;;
let ramPut v n segment =                     (* Wildly inefficient! *)
  let (left, right) = split segment n in
    left @ (v :: right);;

(* Sample SVM programs. Program img is typically executed on svm 
 * with the call: (svm img 0)
 *
 * Let M = data1[0], N = data1[1]. Program1 computes R0 := M * N.
 *)
let data1 = [4; 5];;
let text1 : instruction list = 
  [ 
    Mov {rd=R0; rs=Zero};
    Mov {rd=R3; rs=Zero};
    Lod {rd=R1; offset=0; rs=R3};     (* R1 := M *)
    Lod {rd=R2; offset=1; rs=R3};     (* R2 := N *)
    Li  {rd=R3; number=1};            (* R3 is for decrementing. *)
    Cmp {rs=R2; rt=Zero};
    Beq 3;
    Add {rd=R0; rs=R0; rt=R1};
    Sub {rd=R2; rs=R2; rt=R3};
    Jmp (-5);
    Hlt
  ];;
let image1 : image = {data=data1; text=text1};;

(* program2 counts the data. The value -1 is a sentinal.
 *)
let data2 = [2; 2; 2; 2; 4; 5; 6; 7; -1]
let text2 =
  [
    Mov {rd=R0; rs=Zero};             (* R0 is the counter *)
    Mov {rd=R1; rs=R0};               (* R1 is the address of a number *)
    Li  {rd=R2; number=1};            (* R2 is for incrementing *)
    Lod {rd=R3; offset=0; rs=R1};     (* R3 is the number in the data *)
    Cmp {rs=R3; rt=Zero};            
    Blt 3;
    Add {rd=R1; rs=R1; rt=R2};
    Add {rd=R0; rs=R0; rt=R2};
    Jmp (-6);
    Hlt
  ];;
let image2= {data=data2; text=text2}

(* program3 computes the integer average of the data. NB that position
 * 0 of the data segment is used as a place to store an intermediate
 * result. The data runs from the 2 in position 1 through the sentinal -1.
 *)
let data3 = [0; 30; 40; 50; 60; 70; -1]
let text3 =
  [
    Jsr 13;                           (* R0 := count(the data items) *)
    Mov {rd=R1; rs=Zero};             (* R1 contains the address of data *)
    Sto {rs=R0; offset=0; rt=R1};     (* save # of elements in data area *)
    Mov {rd=R0; rs=R1};               (* R0 now used to hold sum *)
    Li  {rd=R2; number=1};            (* R2 used for incrementing *)
    Lod {rd=R3; offset=1; rs=R1};     (* R3 has the data, NB offset=1! *)
    Cmp {rs=R3; rt=Zero};
    Blt 3;
    Add {rd=R0; rs=R0; rt=R3};        (* R0 := R0 + data *)
    Add {rd=R1; rs=R1; rt=R2};        (* increment the address *)
    Jmp (-6);                         (* go read another data value *)
    Lod {rd=R1; offset=0; rs=Zero};   (* R1 := count *)
    Div {rd=R0; rs=R0; rt=R1};        (* R0 := sum / count, i.e., average *)
    Hlt;
    
    Mov {rd=R0; rs=Zero};             (* start of counting routine *)
    Mov {rd=R1; rs=R0};               (* R1 is the address of a number *)
    Li  {rd=R2; number=1};            (* R2 is for incrementing *)
    Lod {rd=R3; offset=1; rs=R1};     (* R3 is the number in the data *)
    Cmp {rs=R3; rt=Zero};            
    Blt 3;
    Add {rd=R1; rs=R1; rt=R2};
    Add {rd=R0; rs=R0; rt=R2};
    Jmp (-6);
    R    
  ];;
let image3 = {data=data3; text=text3};;

(* Debugging support. 
 *)
let debug = true;;
let step = false;;

let printState pc psw ra registers =
  let regstr = string_of_regs registers 
  in
  begin
    Printf.printf "\ndbg: (pc, psw, ra) = (%d, %d, %d)\n" pc psw ra;
    Printf.printf "dbg: regs = %s\n" regstr
  end

let dbgOut pc psw ra registers =
  match (debug, step) with
  | (true, true) ->
	   begin
	     printState pc psw ra registers;
	     let _ = read_line() in ()
	   end
  | (true, false) -> printState pc psw ra registers
  | _ -> ()

(* An implementation of the Simple Virtual Machine. A call (svm ram pc)
 * executes the SVM program starting with instruction in the text segment
 * of image at pc.
 *)
let svm img pc =

  (* The CPU Instruction cycle.
   *)
  let rec cycle pc psw ra registers ({data; text} as img) =

    (* Fetch the next instruction from the text segment of the image:
     *)
    let instruction = ramGet pc text in
    let newPC = pc + 1 in
    let _ = dbgOut pc psw ra registers
    in
    (* Now dispatch on the instruction.
     *)
    match instruction with
    | Lod {rd; offset; rs} ->
       let addr = offset + (registerGet rs registers) in
	     let value = ramGet addr data in
	     let newRegisters = registerPut value rd registers 
       in
       cycle newPC psw ra newRegisters img

    | Li {rd; number} ->
       let newRegisters = registerPut number rd registers
       in
       cycle newPC psw ra newRegisters img

	  | Sto {rs; offset; rt} ->
       let addr = offset + (registerGet rt registers) in
	     let value = registerGet rs registers in
	     let newData = ramPut value addr data 
       in
       cycle newPC psw ra registers { img with data = newData }

	  | Mov {rd; rs} ->
	     let v = registerGet rs registers in
	     let newRegisters = registerPut v rd registers 
       in
	     cycle newPC psw ra newRegisters img

	  | Add {rd; rs; rt} -> 
	     let v1 = registerGet rs registers in
	     let v2 = registerGet rt registers in
	     let newRegisters = registerPut (v1 + v2) rd registers 
       in
       cycle newPC psw ra newRegisters img

	  | Sub {rd; rs; rt} ->
	     let v1 = registerGet rs registers in
	     let v2 = registerGet rt registers in
	     let newRegisters = registerPut (v1 - v2) rd registers 
       in
       cycle newPC psw ra newRegisters img

	  | Mul {rd; rs; rt} ->
	     let v1 = registerGet rs registers in
	     let v2 = registerGet rt registers in
	     let newRegisters = registerPut (v1 * v2) rd registers 
       in
       cycle newPC psw ra newRegisters img

	  | Div {rd; rs; rt} ->
	     let v1 = registerGet rs registers in
	     let v2 = registerGet rt registers in
	     let newRegisters = registerPut (v1 / v2) rd registers 
       in
       cycle newPC psw ra newRegisters img

	  | Cmp {rs; rt} ->
       let v1 = registerGet rs registers in
	     let v2 = registerGet rt registers in
	     let value = v1 - v2 
       in
	     cycle newPC value ra registers img
(*
	  | Cpz reg -> 
	     let value = registerGet reg registers 
       in
	     cycle newPC value ra registers img
 *)
	  | Blt displacement ->
	     (match psw < 0 with
		    | true  -> cycle (newPC + displacement) psw ra registers img
	      | false -> cycle newPC psw ra registers img)
         
	  | Beq displacement ->
	     (match psw = 0 with
		    | true  -> cycle (newPC + displacement) psw ra registers img
	      | false -> cycle newPC psw ra registers img)

	  | Bgt displacement ->
	     (match psw > 0 with
		    | true  -> cycle (newPC + displacement) psw ra registers img
	      | false -> cycle newPC psw ra registers img)
         
	  | Jmp disp -> cycle (newPC + disp) psw ra registers img
                        
    | Jsr disp -> cycle (newPC + disp) psw newPC registers img
                        
    | R -> cycle ra psw ra registers img

	  | Hlt -> 
	     begin
         Printf.printf "\nSVM Halt.";
	       printState pc psw ra registers
	     end
  in
  let initialPSW = 0 in
  let initialRA = 0 in
  let initialRegisters = {r0=0; r1=0; r2=0; r3=0}
  in
  cycle pc initialPSW initialRA initialRegisters img
